diff --git a/lisp/emacs-lisp/macroexp.el b/lisp/emacs-lisp/macroexp.el
index 6d5cf8723f9..18eaf005cb4 100644
--- a/lisp/emacs-lisp/macroexp.el
+++ b/lisp/emacs-lisp/macroexp.el
@@ -199,6 +199,15 @@ macroexp--obsolete-warning
            (instead (format-message "; use `%s' instead." instead))
            (t ".")))))
 
+(defvar macroexp--whole-form
+  ;; The current form being macroexpanded.
+  )
+
+(defun macroexp-whole-form ()
+  "The current form being macro-expanded.
+Can only be called during macro-expansion (ie, inside a macro)."
+  macroexp--whole-form)
+
 (defun macroexpand-1 (form &optional environment)
   "Perform (at most) one step of macroexpansion."
   (cond
@@ -219,7 +228,8 @@ macroexpand-1
              ((not (consp def)) form)
              (t
               (if (eq 'macro (car def))
-                  (apply (cdr def) (cdr form))
+                  (let ((macroexp--whole-form form))
+                    (apply (cdr def) (cdr form)))
                 form))))))))
    (t form)))
 
diff --git a/src/eval.c b/src/eval.c
index 9e54d489a3b..0d32b63bf1d 100644
--- a/src/eval.c
+++ b/src/eval.c
@@ -1161,7 +1161,9 @@ DEFUN ("macroexpand", Fmacroexpand, Smacroexpand, 1, 2, 0,
 	    break;
 	}
       {
-	Lisp_Object newform = apply1 (expander, XCDR (form));
+	specpdl_ref pdl = SPECPDL_INDEX ();
+	specbind (Qmacroexp__whole_form, form);
+	Lisp_Object newform = unbind_to (pdl, apply1 (expander, XCDR (form)));
 	if (EQ (form, newform))
 	  break;
 	else
@@ -4462,4 +4464,5 @@ syms_of_eval (void)
   defsubr (&Sspecial_variable_p);
   DEFSYM (Qfunctionp, "functionp");
   defsubr (&Sfunctionp);
+  DEFSYM (Qmacroexp__whole_form, "macroexp--whole-form");
 }
